@@ -3,8 +3,6 @@
 Wrap libsodium routines
 '''
 # pylint: disable=C0103
-# Import libnacl libs
-from libnacl.version import __version__
 # Import python libs
 import ctypes
 import sys
@@ -80,7 +78,13 @@ def _get_nacl():
             msg += 'libsodium.so.{0}, '.format(soname_ver)
         raise OSError(msg)
 
-nacl = _get_nacl()
+# Don't load libnacl if we are in sphinx
+if not 'sphinx' in sys.argv[0]:
+    nacl = _get_nacl()
+    DOC_RUN = False
+else:
+    nacl = None
+    DOC_RUN = True
 
 
 # Define exceptions
@@ -89,74 +93,101 @@ class CryptError(Exception):
     Base Exception for cryptographic errors
     """
 
-sodium_init = nacl.sodium_init
-sodium_init.res_type = ctypes.c_int
-if sodium_init() < 0:
-    raise RuntimeError('sodium_init() call failed!')
+if not DOC_RUN:
+    sodium_init = nacl.sodium_init
+    sodium_init.res_type = ctypes.c_int
+    if sodium_init() < 0:
+        raise RuntimeError('sodium_init() call failed!')
 
-# Define constants
-try:
-    crypto_box_SEALBYTES = nacl.crypto_box_sealbytes()
-    HAS_SEAL = True
-except AttributeError:
-    HAS_SEAL = False
-try:
-    crypto_aead_aes256gcm_KEYBYTES = nacl.crypto_aead_aes256gcm_keybytes()
-    crypto_aead_aes256gcm_NPUBBYTES = nacl.crypto_aead_aes256gcm_npubbytes()
-    crypto_aead_aes256gcm_ABYTES = nacl.crypto_aead_aes256gcm_abytes()
-    HAS_AEAD_AES256GCM = bool(nacl.crypto_aead_aes256gcm_is_available())
-    crypto_aead_chacha20poly1305_ietf_KEYBYTES = nacl.crypto_aead_chacha20poly1305_ietf_keybytes()
-    crypto_aead_chacha20poly1305_ietf_NPUBBYTES = nacl.crypto_aead_chacha20poly1305_ietf_npubbytes()
-    crypto_aead_chacha20poly1305_ietf_ABYTES = nacl.crypto_aead_chacha20poly1305_ietf_abytes()
-    HAS_AEAD_CHACHA20POLY1305_IETF = True
-    HAS_AEAD = True
-except AttributeError:
-    HAS_AEAD_AES256GCM = False
-    HAS_AEAD_CHACHA20POLY1305_IETF = False
-    HAS_AEAD = False
+    # Define constants
+    try:
+        crypto_box_SEALBYTES = nacl.crypto_box_sealbytes()
+        HAS_SEAL = True
+    except AttributeError:
+        HAS_SEAL = False
+    try:
+        crypto_aead_aes256gcm_KEYBYTES = nacl.crypto_aead_aes256gcm_keybytes()
+        crypto_aead_aes256gcm_NPUBBYTES = nacl.crypto_aead_aes256gcm_npubbytes()
+        crypto_aead_aes256gcm_ABYTES = nacl.crypto_aead_aes256gcm_abytes()
+        HAS_AEAD_AES256GCM = bool(nacl.crypto_aead_aes256gcm_is_available())
+        crypto_aead_chacha20poly1305_ietf_KEYBYTES = nacl.crypto_aead_chacha20poly1305_ietf_keybytes()
+        crypto_aead_chacha20poly1305_ietf_NPUBBYTES = nacl.crypto_aead_chacha20poly1305_ietf_npubbytes()
+        crypto_aead_chacha20poly1305_ietf_ABYTES = nacl.crypto_aead_chacha20poly1305_ietf_abytes()
+        HAS_AEAD_CHACHA20POLY1305_IETF = True
+        HAS_AEAD = True
+    except AttributeError:
+        HAS_AEAD_AES256GCM = False
+        HAS_AEAD_CHACHA20POLY1305_IETF = False
+        HAS_AEAD = False
 
-crypto_box_SECRETKEYBYTES = nacl.crypto_box_secretkeybytes()
-crypto_box_SEEDBYTES = nacl.crypto_box_seedbytes()
-crypto_box_PUBLICKEYBYTES = nacl.crypto_box_publickeybytes()
-crypto_box_NONCEBYTES = nacl.crypto_box_noncebytes()
-crypto_box_ZEROBYTES = nacl.crypto_box_zerobytes()
-crypto_box_BOXZEROBYTES = nacl.crypto_box_boxzerobytes()
-crypto_box_BEFORENMBYTES = nacl.crypto_box_beforenmbytes()
-crypto_scalarmult_BYTES = nacl.crypto_scalarmult_bytes()
-crypto_scalarmult_SCALARBYTES = nacl.crypto_scalarmult_scalarbytes()
-crypto_sign_BYTES = nacl.crypto_sign_bytes()
-crypto_sign_SEEDBYTES = nacl.crypto_sign_secretkeybytes() // 2
-crypto_sign_PUBLICKEYBYTES = nacl.crypto_sign_publickeybytes()
-crypto_sign_SECRETKEYBYTES = nacl.crypto_sign_secretkeybytes()
-crypto_sign_ed25519_PUBLICKEYBYTES = nacl.crypto_sign_ed25519_publickeybytes()
-crypto_sign_ed25519_SECRETKEYBYTES = nacl.crypto_sign_ed25519_secretkeybytes()
-crypto_box_MACBYTES = crypto_box_ZEROBYTES - crypto_box_BOXZEROBYTES
-crypto_secretbox_KEYBYTES = nacl.crypto_secretbox_keybytes()
-crypto_secretbox_NONCEBYTES = nacl.crypto_secretbox_noncebytes()
-crypto_secretbox_ZEROBYTES = nacl.crypto_secretbox_zerobytes()
-crypto_secretbox_BOXZEROBYTES = nacl.crypto_secretbox_boxzerobytes()
-crypto_secretbox_MACBYTES = crypto_secretbox_ZEROBYTES - crypto_secretbox_BOXZEROBYTES
-crypto_stream_KEYBYTES = nacl.crypto_stream_keybytes()
-crypto_stream_NONCEBYTES = nacl.crypto_stream_noncebytes()
-crypto_auth_BYTES = nacl.crypto_auth_bytes()
-crypto_auth_KEYBYTES = nacl.crypto_auth_keybytes()
-crypto_onetimeauth_BYTES = nacl.crypto_onetimeauth_bytes()
-crypto_onetimeauth_KEYBYTES = nacl.crypto_onetimeauth_keybytes()
-crypto_generichash_BYTES = nacl.crypto_generichash_bytes()
-crypto_generichash_BYTES_MIN = nacl.crypto_generichash_bytes_min()
-crypto_generichash_BYTES_MAX = nacl.crypto_generichash_bytes_max()
-crypto_generichash_KEYBYTES = nacl.crypto_generichash_keybytes()
-crypto_generichash_KEYBYTES_MIN = nacl.crypto_generichash_keybytes_min()
-crypto_generichash_KEYBYTES_MAX = nacl.crypto_generichash_keybytes_max()
-crypto_scalarmult_curve25519_BYTES = nacl.crypto_scalarmult_curve25519_bytes()
-crypto_hash_BYTES = nacl.crypto_hash_sha512_bytes()
-crypto_hash_sha256_BYTES = nacl.crypto_hash_sha256_bytes()
-crypto_hash_sha512_BYTES = nacl.crypto_hash_sha512_bytes()
-crypto_verify_16_BYTES = nacl.crypto_verify_16_bytes()
-crypto_verify_32_BYTES = nacl.crypto_verify_32_bytes()
-crypto_verify_64_BYTES = nacl.crypto_verify_64_bytes()
-# pylint: enable=C0103
+    crypto_box_SECRETKEYBYTES = nacl.crypto_box_secretkeybytes()
+    crypto_box_SEEDBYTES = nacl.crypto_box_seedbytes()
+    crypto_box_PUBLICKEYBYTES = nacl.crypto_box_publickeybytes()
+    crypto_box_NONCEBYTES = nacl.crypto_box_noncebytes()
+    crypto_box_ZEROBYTES = nacl.crypto_box_zerobytes()
+    crypto_box_BOXZEROBYTES = nacl.crypto_box_boxzerobytes()
+    crypto_box_BEFORENMBYTES = nacl.crypto_box_beforenmbytes()
+    crypto_scalarmult_BYTES = nacl.crypto_scalarmult_bytes()
+    crypto_scalarmult_SCALARBYTES = nacl.crypto_scalarmult_scalarbytes()
+    crypto_sign_BYTES = nacl.crypto_sign_bytes()
+    crypto_sign_SEEDBYTES = nacl.crypto_sign_secretkeybytes() // 2
+    crypto_sign_PUBLICKEYBYTES = nacl.crypto_sign_publickeybytes()
+    crypto_sign_SECRETKEYBYTES = nacl.crypto_sign_secretkeybytes()
+    crypto_sign_ed25519_PUBLICKEYBYTES = nacl.crypto_sign_ed25519_publickeybytes()
+    crypto_sign_ed25519_SECRETKEYBYTES = nacl.crypto_sign_ed25519_secretkeybytes()
+    crypto_box_MACBYTES = crypto_box_ZEROBYTES - crypto_box_BOXZEROBYTES
+    crypto_secretbox_KEYBYTES = nacl.crypto_secretbox_keybytes()
+    crypto_secretbox_NONCEBYTES = nacl.crypto_secretbox_noncebytes()
+    crypto_secretbox_ZEROBYTES = nacl.crypto_secretbox_zerobytes()
+    crypto_secretbox_BOXZEROBYTES = nacl.crypto_secretbox_boxzerobytes()
+    crypto_secretbox_MACBYTES = crypto_secretbox_ZEROBYTES - crypto_secretbox_BOXZEROBYTES
+    crypto_stream_KEYBYTES = nacl.crypto_stream_keybytes()
+    crypto_stream_NONCEBYTES = nacl.crypto_stream_noncebytes()
+    crypto_auth_BYTES = nacl.crypto_auth_bytes()
+    crypto_auth_KEYBYTES = nacl.crypto_auth_keybytes()
+    crypto_onetimeauth_BYTES = nacl.crypto_onetimeauth_bytes()
+    crypto_onetimeauth_KEYBYTES = nacl.crypto_onetimeauth_keybytes()
+    crypto_generichash_BYTES = nacl.crypto_generichash_bytes()
+    crypto_generichash_BYTES_MIN = nacl.crypto_generichash_bytes_min()
+    crypto_generichash_BYTES_MAX = nacl.crypto_generichash_bytes_max()
+    crypto_generichash_KEYBYTES = nacl.crypto_generichash_keybytes()
+    crypto_generichash_KEYBYTES_MIN = nacl.crypto_generichash_keybytes_min()
+    crypto_generichash_KEYBYTES_MAX = nacl.crypto_generichash_keybytes_max()
+    crypto_scalarmult_curve25519_BYTES = nacl.crypto_scalarmult_curve25519_bytes()
+    crypto_hash_BYTES = nacl.crypto_hash_sha512_bytes()
+    crypto_hash_sha256_BYTES = nacl.crypto_hash_sha256_bytes()
+    crypto_hash_sha512_BYTES = nacl.crypto_hash_sha512_bytes()
+    crypto_verify_16_BYTES = nacl.crypto_verify_16_bytes()
+    crypto_verify_32_BYTES = nacl.crypto_verify_32_bytes()
+    crypto_verify_64_BYTES = nacl.crypto_verify_64_bytes()
 
+    try:
+        randombytes_SEEDBYTES = nacl.randombytes_seedbytes()
+        HAS_RAND_SEED = True
+    except AttributeError:
+        HAS_RAND_SEED = False
+
+    try:
+        crypto_kdf_PRIMITIVE = nacl.crypto_kdf_primitive()
+        crypto_kdf_BYTES_MIN = nacl.crypto_kdf_bytes_min()
+        crypto_kdf_BYTES_MAX = nacl.crypto_kdf_bytes_max()
+        crypto_kdf_CONTEXTBYTES = nacl.crypto_kdf_contextbytes()
+        crypto_kdf_KEYBYTES = nacl.crypto_kdf_keybytes()
+        HAS_CRYPT_KDF = True
+    except AttributeError:
+        HAS_CRYPT_KDF = False
+
+    try:
+        crypto_kx_PUBLICKEYBYTES = nacl.crypto_kx_publickeybytes()
+        crypto_kx_SECRETKEYBYTES = nacl.crypto_kx_secretkeybytes()
+        crypto_kx_SEEDBYTES = nacl.crypto_kx_seedbytes()
+        crypto_kx_SESSIONKEYBYTES = nacl.crypto_kx_sessionkeybytes()
+        crypto_kx_PRIMITIVE = nacl.crypto_kx_primitive()
+        HAS_CRYPT_KX = True
+    except AttributeError:
+        HAS_CRYPT_KX = False
+
+    # pylint: enable=C0103
 
 # Pubkey defs
 
@@ -187,15 +218,19 @@ def crypto_box_seed_keypair(seed):
 
 def crypto_scalarmult_base(sk):
     '''
-    Generate a public key from a secret key 
+    Compute and return the scalar product of a standard group element and the given integer.
+
+    This can be used to derive a Curve25519 public key from a Curve25519 secret key,
+    such as for usage with crypto_box and crypto_box_seal.
     '''
     if len(sk) != crypto_box_SECRETKEYBYTES:
         raise ValueError('Invalid secret key')
     pk = ctypes.create_string_buffer(crypto_box_PUBLICKEYBYTES)
-    nacl.crypto_scalarmult_base(pk, sk)
+    if nacl.crypto_scalarmult_base(pk, sk):
+        raise CryptError('Failed to compute scalar product')
     return pk.raw
-    
-    
+
+
 def crypto_box(msg, nonce, pk, sk):
     '''
     Using a public key and a secret key encrypt the given message. A nonce
@@ -444,6 +479,9 @@ def crypto_sign_ed25519_sk_to_pk(sk):
     '''
     Extract the public key from the secret key
     '''
+    if len(sk) != crypto_sign_ed25519_SECRETKEYBYTES:
+        raise ValueError('Invalid secret key')
+
     pk = ctypes.create_string_buffer(crypto_sign_PUBLICKEYBYTES)
     ret = nacl.crypto_sign_ed25519_sk_to_pk(pk, sk)
     if ret:
@@ -455,6 +493,9 @@ def crypto_sign_ed25519_sk_to_seed(sk):
     '''
     Extract the seed from the secret key 
     '''
+    if len(sk) != crypto_sign_ed25519_SECRETKEYBYTES:
+        raise ValueError('Invalid secret key')
+
     seed = ctypes.create_string_buffer(crypto_sign_SEEDBYTES)
     ret = nacl.crypto_sign_ed25519_sk_to_seed(seed, sk)
     if ret:
@@ -466,6 +507,9 @@ def crypto_sign(msg, sk):
     '''
     Sign the given message with the given signing key
     '''
+    if len(sk) != crypto_sign_SECRETKEYBYTES:
+        raise ValueError('Invalid secret key')
+
     sig = ctypes.create_string_buffer(len(msg) + crypto_sign_BYTES)
     slen = ctypes.pointer(ctypes.c_ulonglong())
     ret = nacl.crypto_sign(
@@ -483,6 +527,9 @@ def crypto_sign_detached(msg, sk):
     '''
     Return signature for the given message with the given signing key
     '''
+    if len(sk) != crypto_sign_SECRETKEYBYTES:
+        raise ValueError('Invalid secret key')
+
     sig = ctypes.create_string_buffer(crypto_sign_BYTES)
     slen = ctypes.pointer(ctypes.c_ulonglong())
     ret = nacl.crypto_sign_detached(
@@ -502,6 +549,7 @@ def crypto_sign_seed_keypair(seed):
     '''
     if len(seed) != crypto_sign_SEEDBYTES:
         raise ValueError('Invalid Seed')
+
     sk = ctypes.create_string_buffer(crypto_sign_SECRETKEYBYTES)
     vk = ctypes.create_string_buffer(crypto_sign_PUBLICKEYBYTES)
 
@@ -515,6 +563,9 @@ def crypto_sign_open(sig, vk):
     '''
     Verifies the signed message sig using the signer's verification key
     '''
+    if len(vk) != crypto_sign_PUBLICKEYBYTES:
+        raise ValueError('Invalid public key')
+
     msg = ctypes.create_string_buffer(len(sig))
     msglen = ctypes.c_ulonglong()
     msglenp = ctypes.pointer(msglen)
@@ -533,6 +584,11 @@ def crypto_sign_verify_detached(sig, msg, vk):
     '''
     Verifies that sig is a valid signature for the message msg using the signer's verification key
     '''
+    if len(sig) != crypto_sign_BYTES:
+        raise ValueError('Invalid signature')
+    if len(vk) != crypto_sign_PUBLICKEYBYTES:
+        raise ValueError('Invalid public key')
+
     ret = nacl.crypto_sign_verify_detached(
             sig,
             msg,
@@ -600,6 +656,37 @@ def crypto_secretbox_open(ctxt, nonce, key):
         raise ValueError('Failed to decrypt message')
     return msg.raw[crypto_secretbox_ZEROBYTES:]
 
+# Authenticated Symmetric Encryption improved version
+
+
+def crypto_secretbox_easy(cmessage, nonce, key):
+    if len(key) != crypto_secretbox_KEYBYTES:
+        raise ValueError('Invalid key')
+
+    if len(nonce) != crypto_secretbox_NONCEBYTES:
+        raise ValueError('Invalid nonce')
+
+    
+    ctxt = ctypes.create_string_buffer(crypto_secretbox_MACBYTES + len(cmessage))
+    ret = nacl.crypto_secretbox_easy(ctxt, cmessage, ctypes.c_ulonglong(len(cmessage)), nonce, key)
+    if ret:
+        raise ValueError('Failed to encrypt message')
+    return ctxt.raw[0:]
+
+def crypto_secretbox_open_easy(ctxt, nonce, key):
+
+    if len(key) != crypto_secretbox_KEYBYTES:
+        raise ValueError('Invalid key')
+
+    if len(nonce) != crypto_secretbox_NONCEBYTES:
+        raise ValueError('Invalid nonce')
+
+    msg = ctypes.create_string_buffer(len(ctxt))
+    ret = nacl.crypto_secretbox_open_easy(msg, ctxt, ctypes.c_ulonglong(len(ctxt)), nonce, key)
+    if ret:
+        raise ValueError('Failed to decrypt message')
+    return msg.raw[0:len(ctxt) - crypto_secretbox_MACBYTES]    
+
 # Authenticated Symmetric Encryption with Additional Data
 
 
@@ -748,6 +835,11 @@ def crypto_stream(slen, nonce, key):
     '''
     Generates a stream using the given secret key and nonce
     '''
+    if len(key) != crypto_stream_KEYBYTES:
+        raise ValueError('Invalid secret key')
+    if len(nonce) != crypto_stream_NONCEBYTES:
+        raise ValueError('Invalid nonce')
+
     stream = ctypes.create_string_buffer(slen)
     ret = nacl.crypto_stream(stream, ctypes.c_ulonglong(slen), nonce, key)
     if ret:
@@ -763,6 +855,11 @@ def crypto_stream_xor(msg, nonce, key):
     plaintext (xor) the output of crypto_stream. Consequently
     crypto_stream_xor can also be used to decrypt
     '''
+    if len(key) != crypto_stream_KEYBYTES:
+        raise ValueError('Invalid secret key')
+    if len(nonce) != crypto_stream_NONCEBYTES:
+        raise ValueError('Invalid nonce')
+
     stream = ctypes.create_string_buffer(len(msg))
     ret = nacl.crypto_stream_xor(
             stream,
@@ -783,6 +880,9 @@ def crypto_auth(msg, key):
     Constructs a one time authentication token for the given message msg
     using a given secret key
     '''
+    if len(key) != crypto_auth_KEYBYTES:
+        raise ValueError('Invalid secret key')
+
     tok = ctypes.create_string_buffer(crypto_auth_BYTES)
     ret = nacl.crypto_auth(tok, msg, ctypes.c_ulonglong(len(msg)), key)
     if ret:
@@ -795,6 +895,11 @@ def crypto_auth_verify(tok, msg, key):
     Verifies that the given authentication token is correct for the given
     message and key
     '''
+    if len(key) != crypto_auth_KEYBYTES:
+        raise ValueError('Invalid secret key')
+    if len(tok) != crypto_auth_BYTES:
+        raise ValueError('Invalid authenticator')
+
     ret = nacl.crypto_auth_verify(tok, msg, ctypes.c_ulonglong(len(msg)), key)
     if ret:
         raise ValueError('Failed to auth msg')
@@ -919,19 +1024,6 @@ def crypto_generichash(msg, key=None):
             ctypes.c_size_t(key_len))
     return hbuf.raw
 
-# scalarmult
-
-
-def crypto_scalarmult_base(n):
-    '''
-    Computes and returns the scalar product of a standard group element and an
-    integer "n".
-    '''
-    buf = ctypes.create_string_buffer(crypto_scalarmult_BYTES)
-    ret = nacl.crypto_scalarmult_base(buf, n)
-    if ret:
-        raise CryptError('Failed to compute scalar product')
-    return buf.raw
 
 # String cmp
 
@@ -946,7 +1038,8 @@ def crypto_verify_16(string1, string2):
     matching prefix of string1 and string2. This often allows for easy
     timing attacks.
     '''
-    return not nacl.crypto_verify_16(string1, string2)
+    a, b, c = (len(string1) >= 16), (len(string2) >= 16), (not nacl.crypto_verify_16(string1, string2))
+    return a & b & c
 
 
 def crypto_verify_32(string1, string2):
@@ -959,7 +1052,8 @@ def crypto_verify_32(string1, string2):
     matching prefix of string1 and string2. This often allows for easy
     timing attacks.
     '''
-    return not nacl.crypto_verify_32(string1, string2)
+    a, b, c = (len(string1) >= 32), (len(string2) >= 32), (not nacl.crypto_verify_32(string1, string2))
+    return a & b & c
 
 
 def crypto_verify_64(string1, string2):
@@ -972,7 +1066,8 @@ def crypto_verify_64(string1, string2):
     matching prefix of string1 and string2. This often allows for easy
     timing attacks.
     '''
-    return not nacl.crypto_verify_64(string1, string2)
+    a, b, c = (len(string1) >= 64), (len(string2) >= 64), (not nacl.crypto_verify_64(string1, string2))
+    return a & b & c
 
 
 def bytes_eq(a, b):
@@ -1017,6 +1112,22 @@ def randombytes_buf(size):
     nacl.randombytes_buf(buf, size)
     return buf.raw
 
+def randombytes_buf_deterministic(size, seed):
+    '''
+    Returns a string of random byles of the given size for a given seed. 
+    For a given seed, this function will always output the same sequence. 
+    Size can be up to 2^70 (256 GB).
+    '''
+
+    if not HAS_RAND_SEED:
+        raise ValueError('Underlying Sodium library does not support randombytes_seedbytes')
+    if len(seed) != randombytes_SEEDBYTES:
+        raise ValueError('Invalid key seed')
+
+    size = int(size)
+    buf = ctypes.create_string_buffer(size)
+    nacl.randombytes_buf_deterministic(buf, size, seed)
+    return buf.raw 
 
 def randombytes_close():
     '''
@@ -1049,6 +1160,85 @@ def randombytes_uniform(upper_bound):
     '''
     return nacl.randombytes_uniform(upper_bound)
 
+# Key derivation API 
+
+def crypto_kdf_keygen():
+    '''
+    Returns a string of random bytes to generate a master key
+    '''
+    if not HAS_CRYPT_KDF:
+        raise ValueError('Underlying Sodium library does not support crypto_kdf_keybytes')
+    size = crypto_kdf_KEYBYTES
+    buf = ctypes.create_string_buffer(size)
+    nacl.crypto_kdf_keygen(buf)
+    return buf.raw 
+
+def crypto_kdf_derive_from_key(subkey_size, subkey_id, context, master_key):
+    '''
+    Returns a subkey generated from a master key for a given subkey_id. 
+    For a given subkey_id, the subkey will always be the same string.
+    '''
+    size = int(subkey_size)
+    buf = ctypes.create_string_buffer(size)
+    nacl.crypto_kdf_derive_from_key(buf, subkey_size, subkey_id, context, master_key)
+    return buf.raw
+
+# Key Exchange API
+
+def crypto_kx_keypair():
+    '''
+    Generate and return a new keypair
+    '''
+    if not HAS_CRYPT_KX:
+        raise ValueError('Underlying Sodium library does not support crypto_kx')
+    pk = ctypes.create_string_buffer(crypto_kx_PUBLICKEYBYTES)
+    sk = ctypes.create_string_buffer(crypto_kx_SECRETKEYBYTES)
+    nacl.crypto_kx_keypair(pk, sk)
+    return pk.raw, sk.raw
+
+def crypto_kx_seed_keypair(seed):
+    '''
+    Generate and return a keypair from a key seed
+    '''
+    if not HAS_CRYPT_KX:
+        raise ValueError('Underlying Sodium library does not support crypto_kx')
+
+    if len(seed) != crypto_kx_SEEDBYTES:
+        raise ValueError('Invalid key seed')
+    pk = ctypes.create_string_buffer(crypto_kx_PUBLICKEYBYTES)
+    sk = ctypes.create_string_buffer(crypto_kx_SECRETKEYBYTES)
+    nacl.crypto_kx_seed_keypair(pk, sk, seed)
+    return pk.raw, sk.raw
+
+def crypto_kx_client_session_keys(client_pk, client_sk, server_pk):
+    '''
+    Computes a pair of shared keys (rx and tx) using the client's public key client_pk, 
+    the client's secret key client_sk and the server's public key server_pk.
+    Status returns 0 on success, or -1 if the server's public key is not acceptable.
+    '''
+    if not HAS_CRYPT_KX:
+        raise ValueError('Underlying Sodium library does not support crypto_kx')
+
+    rx = ctypes.create_string_buffer(crypto_kx_SESSIONKEYBYTES)
+    tx = ctypes.create_string_buffer(crypto_kx_SESSIONKEYBYTES)
+    status = nacl.crypto_kx_client_session_keys(rx, tx, client_pk, client_sk, server_pk)
+    return rx.raw, tx.raw, status
+
+def crypto_kx_server_session_keys(server_pk, server_sk, client_pk):
+    '''
+    Computes a pair of shared keys (rx and tx) using the server's public key server_pk, 
+    the server's secret key server_sk and the client's public key client_pk.
+    Status returns 0 on success, or -1 if the client's public key is not acceptable.
+    '''
+    if not HAS_CRYPT_KX:
+        raise ValueError('Underlying Sodium library does not support crypto_kx')
+
+    rx = ctypes.create_string_buffer(crypto_kx_SESSIONKEYBYTES)
+    tx = ctypes.create_string_buffer(crypto_kx_SESSIONKEYBYTES)
+    status = nacl.crypto_kx_server_session_keys(rx, tx, server_pk, server_sk, client_pk)
+    return rx.raw, tx.raw, status
+
+
 
 # Utility functions
 
@@ -1075,28 +1265,13 @@ def sodium_version_string():
     return func()
 
 
-def crypto_box_seed_keypair(seed):
-    '''
-    Computes and returns the public and secret keys from the given seed
-    '''
-    if len(seed) != crypto_box_SEEDBYTES:
-        raise ValueError('Invalid Seed')
-    pk = ctypes.create_string_buffer(crypto_box_PUBLICKEYBYTES)
-    sk = ctypes.create_string_buffer(crypto_box_SECRETKEYBYTES)
-
-    ret = nacl.crypto_box_seed_keypair(pk, sk, seed)
-    if ret:
-        raise CryptError('Failed to generate keypair from seed')
-    return (pk.raw, sk.raw)
-
-
 def crypto_sign_ed25519_pk_to_curve25519(ed25519_pk):
     '''
     Convert an Ed25519 public key to a Curve25519 public key
     '''
     if len(ed25519_pk) != crypto_sign_ed25519_PUBLICKEYBYTES:
-        raise ValueError('Invalid Ed25519 Key')
-    
+        raise ValueError('Invalid public key')
+
     curve25519_pk = ctypes.create_string_buffer(crypto_scalarmult_curve25519_BYTES)
     ret = nacl.crypto_sign_ed25519_pk_to_curve25519(curve25519_pk, ed25519_pk)
     if ret:
@@ -1109,10 +1284,12 @@ def crypto_sign_ed25519_sk_to_curve25519(ed25519_sk):
     Convert an Ed25519 secret key to a Curve25519 secret key
     '''
     if len(ed25519_sk) != crypto_sign_ed25519_SECRETKEYBYTES:
-        raise ValueError('Invalid Ed25519 Key')
-    
+        raise ValueError('Invalid secret key')
+
     curve25519_sk = ctypes.create_string_buffer(crypto_scalarmult_curve25519_BYTES)
     ret = nacl.crypto_sign_ed25519_sk_to_curve25519(curve25519_sk, ed25519_sk)
     if ret:
         raise CryptError('Failed to generate Curve25519 secret key')
     return curve25519_sk.raw
+
+
